@@ -79,9 +79,11 @@ static void message(const char *prefix, const char *fmt, va_list *arglist)
 	free(buf);
 }
 
+static int warned = 0;
 static void warn(const char *fmt, ...)
 {
 	va_list arglist;
+	warned++;
 	va_start(arglist, fmt);
 	message("WARNING: ", fmt, &arglist);
 	va_end(arglist);
@@ -117,7 +119,7 @@ static void *do_nofail(void *ptr, const char *file, int line, const char *expr)
 static void print_usage(const char *progname)
 {
 	fprintf(stderr,
-		"Usage: %s [-v] [-V] [-C config-file] [-n] [-i] [-q] [-o <modname>] <modname> [parameters...]\n"
+		"Usage: %s [-v] [-V] [-C config-file] [-n] [-i] [-q] [-b] [-o <modname>] <modname> [parameters...]\n"
 		"%s -r [-n] [-i] [-v] <modulename> ...\n"
 		"%s -l -t <dirname> [ -a <modulename> ...]\n",
 		progname, progname, progname);
@@ -356,11 +358,12 @@ static const char *remove_moderror(int err)
 /* Is module in /proc/modules?  If so, fill in usecount if not NULL. 
    0 means no, 1 means yes, -1 means unknown.
  */
-static int module_in_kernel(const char *modname, int *usecount)
+static int module_in_kernel(const char *modname, unsigned int *usecount)
 {
 	FILE *proc_modules;
 	char *line;
 
+again:
 	/* Might not be mounted yet.  Don't fail. */
 	proc_modules = fopen("/proc/modules", "r");
 	if (!proc_modules)
@@ -369,14 +372,32 @@ static int module_in_kernel(const char *modname, int *usecount)
 	while ((line = getline_wrapped(proc_modules, NULL)) != NULL) {
 		char *entry = strtok(line, " \n");
 
-		if (entry && strcmp(entry, modname) == 0) {
+		if (entry && streq(entry, modname)) {
 			/* If it exists, usecount is the third entry. */
-			if (usecount) {
-				entry = strtok(NULL, " \n");
-				if (entry
-				    && (entry = strtok(NULL, " \n")) != NULL)
+			if (!strtok(NULL, " \n"))
+				goto out;
+
+			if (!(entry = strtok(NULL, " \n"))) /* usecount */
+				goto out;
+			else
+				if (usecount)
 					*usecount = atoi(entry);
+
+			/* Followed by - then status. */
+			if (strtok(NULL, " \n")
+			    && (entry = strtok(NULL, " \n")) != NULL) {
+				/* Locking will fail on ro fs, we might hit
+				 * cases where module is in flux.  Spin. */
+				if (streq(entry, "Loading")
+				    || streq(entry, "Unloading")) {
+					usleep(100000);
+					free(line);
+					fclose(proc_modules);
+					goto again;
+				}
 			}
+
+		out:
 			free(line);
 			fclose(proc_modules);
 			return 1;
@@ -603,6 +624,18 @@ struct module_command
 	char *command;
 };
 
+struct module_alias
+{
+	struct module_alias *next;
+	char *module;
+};
+
+struct module_blacklist
+{
+	struct module_blacklist *next;
+	char *modulename;
+};
+
 /* Link in a new option line from the config file. */
 static struct module_options *
 add_options(const char *modname,
@@ -637,6 +670,57 @@ add_command(const char *modname,
 	return new;
 }
 
+/* Link in a new alias line from the config file. */
+static struct module_alias *
+add_alias(const char *modname, struct module_alias *aliases)
+{
+	struct module_alias *new;
+
+	new = NOFAIL(malloc(sizeof(*new)));
+	new->module = NOFAIL(strdup(modname));
+	new->next = aliases;
+	return new;
+}
+
+/* Link in a new blacklist line from the config file. */
+static struct module_blacklist *
+add_blacklist(const char *modname, struct module_blacklist *blacklist)
+{
+	struct module_blacklist *new;
+
+	new = NOFAIL(malloc(sizeof(*new)));
+	new->modulename = NOFAIL(strdup(modname));
+	new->next = blacklist;
+	return new;
+}
+
+/* Find blacklist commands if any. */
+static  int
+find_blacklist(const char *modname, const struct module_blacklist *blacklist)
+{
+	while (blacklist) {
+		if (strcmp(blacklist->modulename, modname) == 0)
+			return 1;
+		blacklist = blacklist->next;
+	}
+	return 0;
+}
+
+/* return a new alias list, with backlisted elems filtered out */
+static struct module_alias *
+apply_blacklist(const struct module_alias *aliases,
+		const struct module_blacklist *blacklist)
+{
+	struct module_alias *result = NULL;
+	while (aliases) {
+		char *modname = aliases->module;
+		if (!find_blacklist(modname, blacklist))
+			result = add_alias(modname, result);
+		aliases = aliases->next;
+	}
+	return result;
+}
+
 /* Find install commands if any. */
 static const char *find_command(const char *modname,
 				const struct module_command *commands)
@@ -684,23 +768,36 @@ static void verbose_printf(int verbose, const char *fmt, ...)
 	}
 }
 
-/* Do an install/remove command. */
+/* Do an install/remove command: replace $CMDLINE_OPTS if it's specified. */
 static void do_command(const char *modname,
 		       const char *command,
 		       int verbose, int dry_run,
 		       errfn_t error,
-		       const char *type)
+		       const char *type,
+		       const char *cmdline_opts)
 {
 	int ret;
+	char *p, *replaced_cmd = NOFAIL(strdup(command));
 
-	verbose_printf(verbose, "%s %s\n", type, command);
+	while ((p = strstr(replaced_cmd, "$CMDLINE_OPTS")) != NULL) {
+		char *new;
+		asprintf(&new, "%.*s%s%s",
+			 p - replaced_cmd, replaced_cmd, cmdline_opts,
+			 p + strlen("$CMDLINE_OPTS"));
+		NOFAIL(new);
+		free(replaced_cmd);
+		replaced_cmd = new;
+	}
+
+	verbose_printf(verbose, "%s %s\n", type, replaced_cmd);
 	if (dry_run)
 		return;
 
 	setenv("MODPROBE_MODULE", modname, 1);
-	ret = system(command);
+	ret = system(replaced_cmd);
 	if (ret == -1 || WEXITSTATUS(ret))
 		error("Error running %s command for %s\n", type, modname);
+	free(replaced_cmd);
 }
 
 /* Actually do the insert.  Frees second arg. */
@@ -716,7 +813,8 @@ static void insmod(struct list_head *list,
 		   int ignore_commands,
 		   int ignore_proc,
 		   int strip_vermagic,
-		   int strip_modversion)
+		   int strip_modversion,
+		   const char *cmdline_opts)
 {
 	int ret, fd;
 	unsigned long len;
@@ -732,9 +830,10 @@ static void insmod(struct list_head *list,
 	if (!list_empty(list)) {
 		insmod(list, NOFAIL(strdup("")), NULL, 0, warn,
 		       dry_run, verbose, options, commands, 0, ignore_proc,
-		       strip_vermagic, strip_modversion);
+		       strip_vermagic, strip_modversion, cmdline_opts);
 	}
 
+	/* Lock before we look, in case it's initializing. */
 	fd = lock_file(mod->filename);
 	if (fd < 0) {
 		error("Could not open '%s': %s\n",
@@ -744,15 +843,19 @@ static void insmod(struct list_head *list,
 
 	/* Don't do ANYTHING if already in kernel. */
 	if (!ignore_proc
-	    && module_in_kernel(newname ?: mod->modname, NULL) == 1)
-		goto exists_error;
+	    && module_in_kernel(newname ?: mod->modname, NULL) == 1) {
+		if (first_time)
+			error("Module %s already in kernel.\n",
+			      newname ?: mod->modname);
+		goto out_unlock;
+	}
 
 	command = find_command(mod->modname, commands);
 	if (command && !ignore_commands) {
 		/* It might recurse: unlock. */
 		unlock_file(fd);
 		do_command(mod->modname, command, verbose, dry_run, error,
-			   "install");
+			   "install", cmdline_opts);
 		goto out_optstring;
 	}
 
@@ -782,8 +885,12 @@ static void insmod(struct list_head *list,
 
 	ret = init_module(map, len, optstring);
 	if (ret != 0) {
-		if (errno == EEXIST)
-			goto exists_error;
+		if (errno == EEXIST) {
+			if (first_time)
+				error("Module %s already in kernel.\n",
+				      newname ?: mod->modname);
+			goto out_unlock;
+		}
 		error("Error inserting %s (%s): %s\n",
 		      mod->modname, mod->filename, insert_moderror(errno));
 	}
@@ -794,12 +901,6 @@ static void insmod(struct list_head *list,
  out_optstring:
 	free(optstring);
 	return;
-
-exists_error:
-	if (first_time)
-		error("Module %s already in kernel.\n",
-		      newname ?: mod->modname);
-	goto out_unlock;
 }
 
 /* Do recursive removal. */
@@ -811,7 +912,8 @@ static void rmmod(struct list_head *list,
 		  int verbose,
 		  struct module_command *commands,
 		  int ignore_commands,
-		  int ignore_inuse)
+		  int ignore_inuse,
+		  const char *cmdline_opts)
 {
 	const char *command;
 	unsigned int usecount = 0;
@@ -833,7 +935,7 @@ static void rmmod(struct list_head *list,
 		/* It might recurse: unlock. */
 		unlock_file(lock);
 		do_command(mod->modname, command, verbose, dry_run, error,
-			   "remove");
+			   "remove", cmdline_opts);
 		goto remove_rest_no_unlock;
 	}
 
@@ -865,7 +967,7 @@ static void rmmod(struct list_head *list,
 	/* Now do things we depend. */
 	if (!list_empty(list))
 		rmmod(list, NULL, 0, warn, dry_run, verbose, commands,
-		      0, 1);
+		      0, 1, cmdline_opts);
 	return;
 
 nonexistent_module:
@@ -957,7 +1059,8 @@ static int read_config(const char *filename,
 		       int removing,
 		       struct module_options **options,
 		       struct module_command **commands,
-		       char **alias);
+		       struct module_alias **alias,
+		       struct module_blacklist **blacklist);
 
 /* FIXME: Maybe should be extended to "alias a b [and|or c]...". --RR */
 static int read_config_file(const char *filename,
@@ -966,7 +1069,8 @@ static int read_config_file(const char *filename,
 			    int removing,
 			    struct module_options **options,
 			    struct module_command **commands,
-			    char **alias)
+			    struct module_alias **aliases,
+			    struct module_blacklist **blacklist)
 {
 	char *line;
 	unsigned int linenum = 0;
@@ -996,9 +1100,10 @@ static int read_config_file(const char *filename,
 			if (!wildcard || !realname)
 				grammar(cmd, filename, linenum);
 			else if (fnmatch(wildcard,name,0) == 0)
-				*alias = NOFAIL(strdup(realname));
+				*aliases = add_alias(realname, *aliases);
 		} else if (strcmp(cmd, "include") == 0) {
-			char *newalias = NULL, *newfilename;
+			struct module_alias *newalias = NULL;
+			char *newfilename;
 
 			newfilename = strsep_skipspace(&ptr, "\t ");
 			if (!newfilename)
@@ -1006,7 +1111,8 @@ static int read_config_file(const char *filename,
 			else {
 				if (!read_config(newfilename, name,
 						 dump_only, removing,
-						 options, commands, &newalias))
+						 options, commands, &newalias,
+						 blacklist))
 					warn("Failed to open included"
 					      " config file %s: %s\n",
 					      newfilename, strerror(errno));
@@ -1014,7 +1120,7 @@ static int read_config_file(const char *filename,
 				/* Files included override aliases,
 				   etc that was already set ... */
 				if (newalias)
-					*alias = newalias;
+					*aliases = newalias;
 			}
 		} else if (strcmp(cmd, "options") == 0) {
 			modname = strsep_skipspace(&ptr, "\t ");
@@ -1034,6 +1140,14 @@ static int read_config_file(const char *filename,
 				*commands = add_command(underscores(modname),
 							ptr, *commands);
 			}
+		} else if (strcmp(cmd, "blacklist") == 0) {
+			modname = strsep_skipspace(&ptr, "\t ");
+			if (!modname)
+				grammar(cmd, filename, linenum);
+			else if (!removing) {
+				*blacklist = add_blacklist(underscores(modname),
+							*blacklist);
+			}
 		} else if (strcmp(cmd, "remove") == 0) {
 			modname = strsep_skipspace(&ptr, "\t ");
 			if (!modname || !ptr)
@@ -1053,22 +1167,23 @@ static int read_config_file(const char *filename,
 }
 
 /* Simple format, ignore lines starting with #, one command per line.
-   Returns NULL or resolved alias. */
+   Returns true or false. */
 static int read_config(const char *filename,
 		       const char *name,
 		       int dump_only,
 		       int removing,
 		       struct module_options **options,
 		       struct module_command **commands,
-		       char **alias)
+		       struct module_alias **aliases,
+		       struct module_blacklist **blacklist)
 {
 	DIR *dir;
+	int ret = 0;
 
-	/* If it's a directory, recurse. */
+	/* Reiser4 has file/directory duality: treat it as both. */
 	dir = opendir(filename);
 	if (dir) {
 		struct dirent *i;
-
 		while ((i = readdir(dir)) != NULL) {
 			if (!streq(i->d_name,".") && !streq(i->d_name,"..")) {
 				char sub[strlen(filename) + 1
@@ -1076,19 +1191,22 @@ static int read_config(const char *filename,
 
 				sprintf(sub, "%s/%s", filename, i->d_name);
 				if (!read_config(sub, name,
-						 dump_only, removing,
-						 options, commands, alias))
+						 dump_only, removing, options,
+						 commands, aliases, blacklist))
 					warn("Failed to open"
-					      " config file %s: %s\n",
-					      sub, strerror(errno));
+					     " config file %s: %s\n",
+					     sub, strerror(errno));
 			}
 		}
 		closedir(dir);
-		return 1;
+		ret = 1;
 	}
 
-	return read_config_file(filename, name, dump_only, removing,
-				options, commands, alias);
+	if (read_config_file(filename, name, dump_only, removing,
+			     options, commands, aliases, blacklist))
+		ret = 1;
+
+	return ret;
 }
 
 static const char *default_configs[] = 
@@ -1103,13 +1221,14 @@ static void read_toplevel_config(const char *filename,
 				 int removing,
 				 struct module_options **options,
 				 struct module_command **commands,
-				 char **alias)
+				 struct module_alias **aliases,
+				 struct module_blacklist **blacklist)
 {
 	unsigned int i;
 
 	if (filename) {
 		if (!read_config(filename, name, dump_only, removing,
-				 options, commands, alias))
+				 options, commands, aliases, blacklist))
 			fatal("Failed to open config file %s: %s\n",
 			      filename, strerror(errno));
 		return;
@@ -1117,8 +1236,8 @@ static void read_toplevel_config(const char *filename,
 
 	/* Try defaults. */
 	for (i = 0; i < ARRAY_SIZE(default_configs); i++) {
-		if (read_config(default_configs[i], name, dump_only,
-				removing, options, commands, alias))
+		if (read_config(default_configs[i], name, dump_only, removing,
+				options, commands, aliases, blacklist))
 			return;
 	}
 }
@@ -1184,6 +1303,51 @@ static char *gather_options(char *argv[])
 	return optstring;
 }
 
+static void handle_module(const char *modname,
+			  struct list_head *todo_list,
+			  const char *newname,
+			  int remove,
+			  char *options,
+			  int first_time,
+			  errfn_t error,
+			  int dry_run,
+			  int verbose,
+			  struct module_options *modoptions,
+			  struct module_command *commands,
+			  int ignore_commands,
+			  int ignore_proc,
+			  int strip_vermagic,
+			  int strip_modversion,
+			  int unknown_silent,
+			  const char *cmdline_opts)
+{
+	if (list_empty(todo_list)) {
+		const char *command;
+
+		/* The dependencies have to be real modules, but
+		   handle case where the first is completely bogus. */
+		command = find_command(modname, commands);
+		if (command && !ignore_commands) {
+			do_command(modname, command, verbose, dry_run, error,
+				   remove ? "remove":"install", cmdline_opts);
+			return;
+		}
+
+		if (unknown_silent)
+			exit(1);
+		error("Module %s not found.\n", modname);
+		return;
+	}
+
+	if (remove)
+		rmmod(todo_list, newname, first_time, error, dry_run, verbose,
+		      commands, ignore_commands, 0, cmdline_opts);
+	else
+		insmod(todo_list, NOFAIL(strdup(options)), newname,
+		       first_time, error, dry_run, verbose, modoptions,
+		       commands, ignore_commands, ignore_proc, strip_vermagic,
+		       strip_modversion, cmdline_opts);
+}
 
 static struct option options[] = { { "verbose", 0, NULL, 'v' },
 				   { "version", 0, NULL, 'V' },
@@ -1207,6 +1371,7 @@ static struct option options[] = { { "verbose", 0, NULL, 'v' },
 				   { "set-version", 1, NULL, 'S' },
 				   { "show-depends", 0, NULL, 'D' },
 				   { "first-time", 0, NULL, 3 },
+				   { "use-blacklist", 0, NULL, 'b' },
 				   { NULL, 0, NULL, 0 } };
 
 #define MODPROBE_DEVFSD_CONF "/etc/modprobe.devfs"
@@ -1243,9 +1408,10 @@ int main(int argc, char *argv[])
 	int strip_modversion = 0;
 	int ignore_proc = 0;
 	int first_time = 0;
+	int use_blacklist = 0;
 	unsigned int i, num_modules;
 	char *type = NULL;
-	const char *config = NULL, *command = NULL;
+	const char *config = NULL;
 	char *dirname, *optstring;
 	char *newname = NULL;
 	char *aliasfilename, *symfilename;
@@ -1264,7 +1430,7 @@ int main(int argc, char *argv[])
 		try_old_version("modprobe", argv);
 
 	uname(&buf);
-	while ((opt = getopt_long(argc, argv, "vVC:o:rknqQsclt:aif", options, NULL)) != -1){
+	while ((opt = getopt_long(argc, argv, "vVC:o:rknqQsclt:aifb", options, NULL)) != -1){
 		switch (opt) {
 		case 'v':
 			add_to_env_var("-v");
@@ -1346,6 +1512,9 @@ int main(int argc, char *argv[])
 			strip_vermagic = 1;
 			strip_modversion = 1;
 			break;
+		case 'b':
+			use_blacklist = 1;
+			break;
 		case 1:
 			strip_vermagic = 1;
 			break;
@@ -1391,12 +1560,15 @@ int main(int argc, char *argv[])
 	if (dump_only) {
 		struct module_command *commands = NULL;
 		struct module_options *modoptions = NULL;
-		char *a = NULL;
+		struct module_alias *aliases = NULL;
+		struct module_blacklist *blacklist = NULL;
 
 		read_toplevel_config(config, "", 1, 0,
-				     &modoptions, &commands, &a);
-		read_config(aliasfilename, "", 1, 0,&modoptions, &commands,&a);
-		read_config(symfilename, "", 1, 0, &modoptions, &commands, &a);
+			     &modoptions, &commands, &aliases, &blacklist);
+		read_config(aliasfilename, "", 1, 0,&modoptions, &commands,
+			    &aliases, &blacklist);
+		read_config(symfilename, "", 1, 0, &modoptions, &commands,
+			    &aliases, &blacklist);
 		exit(0);
 	}
 
@@ -1412,77 +1584,77 @@ int main(int argc, char *argv[])
 	for (i = 0; i < num_modules; i++) {
 		struct module_command *commands = NULL;
 		struct module_options *modoptions = NULL;
+		struct module_alias *aliases = NULL;
+		struct module_blacklist *blacklist = NULL;
 		LIST_HEAD(list);
 		char *modulearg = argv[optind + i];
-		char *modname = NULL;
 
 		/* Convert name we are looking for */
 		underscores(modulearg);
 
 		/* Returns the resolved alias, options */
 		read_toplevel_config(config, modulearg, 0,
-				     remove, &modoptions, &commands, &modname);
+		     remove, &modoptions, &commands, &aliases, &blacklist);
 
 		/* No luck?  Try symbol names, if starts with symbol:. */
-		if (!modname
-		    && strncmp(modulearg, "symbol:", strlen("symbol:") == 0))
+		if (!aliases
+		    && strncmp(modulearg, "symbol:", strlen("symbol:")) == 0)
 			read_config(symfilename, modulearg, 0,
-				    remove, &modoptions, &commands, &modname);
+			    remove, &modoptions, &commands,
+			    	&aliases, &blacklist);
 
-		/* If we have an alias, gather any options associated with it
-		   (needs to happen after parsing complete). */
-		if (modname) {
-		got_modname:
-			optstring = add_extra_options(modulearg, optstring,
-						      modoptions);
-			read_depends(dirname, modname, &list);
-		} else {
+		if (!aliases) {
+			/* We only use canned aliases as last resort. */
 			read_depends(dirname, modulearg, &list);
-			/* We don't allow canned aliases to override
-			   real modules, or install commands, so we
-			   delay lookup until now. */
+
 			if (list_empty(&list)
-			    && !find_command(modulearg, commands)
-			    && read_config(aliasfilename,
-					   modulearg, 0,
-					   remove, &modoptions,
-					   &commands, &modname)
-			    && modname)
-				goto got_modname;
-
-			modname = strdup(modulearg);
-		}
-
-		if (list_empty(&list)) {
-			/* The dependencies have to be real modules, but
-			   handle case where the first is completely bogus. */
-			command = find_command(modname, commands);
-			if (command && !ignore_commands) {
-				do_command(modname, command, verbose, dry_run,
-					   fatal, remove ? "remove":"install");
-				continue;
+			    && !find_command(modulearg, commands))
+			{
+				read_config(aliasfilename, modulearg, 0,
+					    remove, &modoptions, &commands,
+					    &aliases, &blacklist);
+				aliases = apply_blacklist(aliases, blacklist);
 			}
-			if (unknown_silent)
-				exit(1);
-			error("Module %s not found.\n", modname);
-			continue;
 		}
 
-		if (remove)
-			rmmod(&list, newname, first_time, error, dry_run,
-			      verbose, commands, ignore_commands, 0);
-		else
-			insmod(&list, NOFAIL(strdup(optstring)), newname,
-			       first_time, error, dry_run, verbose, modoptions,
-			       commands, ignore_commands, ignore_proc,
-			       strip_vermagic, strip_modversion);
-		free(modname);
-	}
+		if (aliases) {
+			errfn_t err = error;
 
-	free(dirname);
-	free(aliasfilename);
-	free(symfilename);
-	free(optstring);
+			/* More than one alias?  Don't bail out on failure. */
+			if (aliases->next)
+				err = warn;
+			while (aliases) {
+				/* Add the options for this alias. */
+				char *opts = NOFAIL(strdup(optstring));
+				opts = add_extra_options(modulearg,
+							 opts, modoptions);
+
+				read_depends(dirname, aliases->module, &list);
+				handle_module(aliases->module, &list, newname,
+					      remove, opts, first_time, err,
+					      dry_run, verbose, modoptions,
+					      commands, ignore_commands,
+					      ignore_proc, strip_vermagic,
+					      strip_modversion,
+					      unknown_silent,
+					      optstring);
+
+				aliases = aliases->next;
+				INIT_LIST_HEAD(&list);
+			}
+		} else {
+			if (use_blacklist
+			    && find_blacklist(modulearg, blacklist))
+				continue;
+
+			handle_module(modulearg, &list, newname, remove,
+				      optstring, first_time, error, dry_run,
+				      verbose, modoptions, commands,
+				      ignore_commands, ignore_proc,
+				      strip_vermagic, strip_modversion,
+				      unknown_silent, optstring);
+		}
+	}
 	if (log)
 		closelog();
 
